import contextlib
import inspect
from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from typing import Any

import pygame
import pygame.locals as pg
from pygame.event import Event

from pygskin.rect import add_padding
from pygskin.rect import get_rect_attrs

CLICKABLE = 1
EDITABLE = 2
FOCUSABLE = 4
SCROLLABLE = 8
HAS_BORDER = 16
HAS_BACKGROUND = 32
HAS_SHADOW = 64


Align = Enum("Align", "left center right")
VAlign = Enum("VAlign", "top middle bottom")


@dataclass
class Widget:
    value: Any = None
    flags: int = 0
    rect: pygame.Rect | None = None
    triggered: bool = False
    type: str | None = None
    id: str | None = None
    classes: list[str] = field(default_factory=list)
    pseudo_classes: set[str] = field(default_factory=set)

    def draw(self, surface: pygame.Surface, **style) -> None:
        if "focus" in self.pseudo_classes:
            draw_widget_focus(surface, self, **style)
        draw_widget_background(surface, self, **style)
        draw_widget_border(surface, self, **style)
        draw_widget_text(surface, self, **style)


def label(text: str | list[str], **kwargs) -> Widget:
    return Widget(text, type="label", **kwargs)


def button(text: str | list[str], **kwargs) -> Widget:
    return Widget(
        text,
        CLICKABLE | FOCUSABLE | HAS_BORDER | HAS_BACKGROUND,
        type="button",
        **kwargs,
    )


def textfield(text: list[str], **kwargs) -> Widget:
    return Widget(
        text,
        EDITABLE | FOCUSABLE | HAS_BORDER | HAS_BACKGROUND,
        type="textfield",
        **kwargs,
    )


@dataclass
class radio(Widget):  # noqa
    flags: int = CLICKABLE | FOCUSABLE | HAS_BACKGROUND

    def draw(
        self,
        surface: pygame.Surface,
        has_focus: bool = False,
        checked: bool = False,
        **style,
    ) -> None:
        if has_focus:
            draw_widget_focus(surface, self, **style)
        draw_widget_background(surface, self, **style)
        if self.rect is not None:
            box = self.rect.move_to(
                width=self.rect.height,
                right=self.rect.x - 5,
                y=self.rect.y - self.rect.height * 0.1,
            )
            pygame.draw.rect(surface, style.get("border_color", "white"), box, 1)
            if checked:
                check = pygame.FRect(box).scale_by(0.5)
                pygame.draw.rect(surface, style.get("border_color", "white"), check)
        draw_widget_text(surface, self, **style)


@dataclass
class IMGUI:
    active: int | None = None
    hot: int | None = None
    focus: int | None = None
    tabindex_prev: int | None = None
    widgets: dict[int, Widget] = field(default_factory=dict)
    events: list[Event] = field(default_factory=list)


@contextlib.contextmanager
def render(ui: IMGUI, surface: pygame.Surface, get_styles=None):
    yield get_renderer(ui, surface, get_styles)
    ui.hot = None
    if not pygame.mouse.get_pressed()[0]:
        ui.active = None
    elif ui.active is None:
        ui.active = -1


def get_renderer(ui: IMGUI, surface: pygame.Surface, get_styles=None):

    def render(widget: Widget, **style) -> Widget:
        frame_info = inspect.stack()[1]
        widget_id = hash((frame_info.filename, frame_info.lineno))
        if callable(get_styles):
            style = get_styles(widget) | style
        with get_widget(ui, widget_id, widget, **style) as _widget:
            _widget.draw(surface, **style)
        return _widget.triggered

    return render


@contextlib.contextmanager
def get_widget(ui: IMGUI, widget_id: int, widget: Widget, **kwargs):
    widget.rect = get_widget_rect(widget, **kwargs)
    widget.triggered = False
    widget.pseudo_classes = set()

    if widget.rect.collidepoint(pygame.mouse.get_pos()):
        ui.hot = widget_id
        widget.pseudo_classes.add("hover")

    if ui.hot == widget_id and pygame.mouse.get_pressed()[0]:
        ui.active = widget_id
        widget.pseudo_classes.add("active")
        if widget.flags & CLICKABLE:
            widget.triggered = True

    if widget.flags & FOCUSABLE and ui.focus is None:
        ui.focus = widget_id
        widget.pseudo_classes.add("focus")

    if ui.focus == widget_id:
        widget.pseudo_classes.add("focus")
        handle_keys(widget, ui)

    yield widget

    if widget.flags & FOCUSABLE:
        ui.tabindex_prev = widget_id


def get_widget_rect(widget: Widget, **kwargs) -> pygame.Rect:
    if not widget.rect:
        widget.rect = pygame.Rect(0, 0, 0, 0)
        if isinstance(widget.value, (str, list)):
            font = kwargs.setdefault(
                "font",
                pygame.font.Font(None, kwargs.get("font_size", 20)),
            )
            widget.rect.size = font.size("".join(widget.value))
        if padding := kwargs.get("padding"):
            widget.rect, _ = add_padding(widget.rect, padding)
    if rect_kwargs := get_rect_attrs(kwargs):
        widget.rect = widget.rect.move_to(**rect_kwargs)
    return widget.rect


def draw_widget_background(surface: pygame.Surface, widget: Widget, **style) -> None:
    if widget.flags & HAS_BACKGROUND and widget.rect:
        color = style.get(
            "background_color",
            "white" if "active" in widget.pseudo_classes else "black",
        )
        pygame.draw.rect(
            surface,
            color,
            widget.rect,
            width=0,
            border_radius=style.get("border_radius", 0),
        )


def draw_widget_border(surface: pygame.Surface, widget: Widget, **style) -> None:
    if widget.flags & HAS_BORDER and widget.rect:
        pygame.draw.rect(
            surface,
            style.get("border_color", "white"),
            widget.rect,
            width=style.get("border_width", 1),
            border_radius=style.get("border_radius", 0),
        )


def draw_widget_focus(surface: pygame.Surface, widget: Widget, **style) -> None:
    if widget.rect:
        pygame.draw.rect(
            surface,
            style.get("focus_border_color", "red"),
            widget.rect.inflate(4, 4),
            width=style.get("focus_border_width", 4),
            border_radius=style.get("focus_border_radius", 0),
        )


def draw_widget_text(surface: pygame.Surface, widget: Widget, **style) -> None:
    if not widget.rect:
        return
    match widget.value:
        case str() | list():
            text = "".join(widget.value)
        case (str(text), _):
            pass
        case _:
            return None
    font = style.setdefault(
        "font",
        pygame.font.Font(None, style.get("font_size", 30)),
    )
    color = style.get("color", "white")
    label = font.render(text, True, color)
    rect = label.get_rect(center=widget.rect.center)
    match Align[style.get("align", "center")]:
        case Align.left:
            rect.left = widget.rect.left
        case Align.right:
            rect.right = widget.rect.right
        case _:
            rect.centerx = widget.rect.centerx
    match VAlign[style.get("valign", "middle")]:
        case VAlign.top:
            rect.top = widget.rect.top
        case VAlign.bottom:
            rect.bottom = widget.rect.bottom
        case _:
            rect.centery = widget.rect.centery
    _, rect = add_padding(rect, style.get("padding", 0))
    surface.blit(label, rect)


def handle_keys(widget: Widget, ui: IMGUI) -> None:
    for event in ui.events:
        if event.type != pg.KEYDOWN:
            continue
        if event.key == pg.K_TAB:
            if ui.focus and ui.widgets.get(ui.focus) == widget:
                ui.focus = None
                if event.mod & pg.KMOD_SHIFT:
                    ui.focus = ui.tabindex_prev
            # prevent this event from being processed by other widgets
            event.key = 0
        if widget.flags & EDITABLE and event.key == pg.K_BACKSPACE:
            widget.value[:] = widget.value[:-1]
            widget.triggered = True
        if widget.flags & EDITABLE and 32 <= event.key < 127 and len(widget.value) < 30:
            widget.value.append(event.unicode)
            widget.triggered = True



import pygame
from pygame import Surface
from pygame.typing import ColorLike
from pygame.typing import Point


def make_color_gradient(
    size: Point,
    start: ColorLike,
    end: ColorLike,
    vertical: bool = True,
    invert: bool = False,
) -> Surface:
    width, height = int(size[0]), int(size[1])

    if vertical:
        positions = zip([0] * height, range(height))
        stripe = Surface((1, height), pygame.SRCALPHA)
        quotient = 1 / height
    else:
        positions = zip(range(width), [0] * width)
        stripe = Surface((width, 1), pygame.SRCALPHA)
        quotient = 1 / width

    stripe.fill((0, 0, 0, 0))

    if invert:
        start, end = end, start

    for i, pos in enumerate(positions):
        stripe.set_at(pos, pygame.Color(start).lerp(pygame.Color(end), i * quotient))

    return pygame.transform.scale(stripe, (width, height))




from collections.abc import Sequence

from pygame import Rect

RECT_ATTRS = """
    x y
    top left bottom right
    topleft bottomleft topright bottomright
    midtop midleft midbottom midright
    center centerx centery
    size width height
    w h
""".strip().split()


def get_rect_attrs(d: dict) -> dict:
    return {k: v for k, v in d.items() if k in RECT_ATTRS}


def add_padding(
    rect: Rect,
    padding: int | Sequence[int] = 0,
) -> tuple[Rect, Rect]:
    match padding:
        case int(px):
            return rect.inflate(px * 2, px * 2), rect.move(px, px)

        case Sequence() as p if 1 <= len(p) <= 4 and all(isinstance(i, int) for i in p):
            try:
                it = iter(p)
                top = right = bottom = left = next(it)
                right = left = next(it)
                bottom = next(it)
                left = next(it)
            except StopIteration:
                pass
            return (
                rect.inflate(left + right, top + bottom).move_to(
                    top=rect.top - top, left=rect.left - left
                ),
                rect.move(left, top),
            )

    raise ValueError("Invalid padding")



from dataclasses import dataclass

import pygame


@dataclass
class Spritesheet:
    """A class to represent a spritesheet.

    >>> _ = pygame.init()
    >>> _ = pygame.display.set_mode((100, 50))
    >>> image = pygame.Surface((100, 50))
    >>> colors = "red yellow green brown blue pink black white".split()
    >>> coords = [(x, y) for y in range(2) for x in range(4)]
    >>> name_map = dict(zip(colors, coords))
    >>> for i, (color, (x, y)) in enumerate(name_map.items()):
    ...     pygame.draw.rect(image, colors[i], (x * 25, y * 25, 25, 25))
    Rect(0, 0, 25, 25)
    ...
    >>> sheet = Spritesheet(image, rows=2, columns=4, name_map=name_map)
    >>> all(
    ...     all(
    ...         sheet[color].get_at((x, y)) == pygame.Color(color)
    ...         for x in range(25) for y in range(25)
    ...     )
    ...     for color in colors
    ... )
    True
    """

    image: pygame.Surface
    rows: int = 1
    columns: int = 1
    name_map: dict[str, tuple[int, int]] | None = None

    def __post_init__(self) -> None:
        self.grid_size = pygame.Vector2(self.columns, self.rows)

    @property
    def cell_size(self) -> tuple[int, int]:
        if not hasattr(self, "_cell_size"):
            width, height = self.image.get_size()
            self._cell_size = (width // self.columns, height // self.rows)
        return self._cell_size

    def __getitem__(self, key: str | tuple[int, int]) -> pygame.Surface:
        match key:
            case str(skey) if self.name_map and (coord := self.name_map[skey]):
                return self[coord]

            case tuple(t) if (0 <= t[0] < self.columns and 0 <= t[1] < self.rows):
                cell_w, cell_h = self.cell_size
                pos = (t[0] * cell_w, t[1] * cell_h)
                return self.image.subsurface((pos, self.cell_size))

        raise KeyError



from __future__ import annotations

import inspect
from pathlib import Path
from typing import Any

import pygame


class Asset:
    """Base class for assets."""

    suffixes: set[str] = set()
    _class_by_suffix: dict[str, type[Asset]] = {}

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        Asset._class_by_suffix.update({suffix: cls for suffix in cls.suffixes})

    @classmethod
    def load(cls, path: Path) -> Any:
        raise NotImplementedError


class Assets(Asset):
    """
    Provides attribute access to static assets in a directory.

    >>> import tempfile
    >>> tempdir = tempfile.TemporaryDirectory()
    >>> Path(f"{tempdir.name}/foo.gif").write_bytes(
    ...     bytes.fromhex('4749463839610100010000ff002c00000000010001000002003b'))
    26
    >>> _ = pygame.init()
    >>> _ = pygame.display.set_mode((1, 1), pygame.HIDDEN)
    >>> Assets(tempdir.name).foo.get_size()
    (1, 1)
    """

    def __init__(self, path: str | Path | None = None) -> None:
        if path is None:
            # Find the path of the calling script.
            path = Path(inspect.stack()[1].filename).parent / "assets"
            if not path.is_file():
                path = Path("assets")
        self.path = Path(path)

    def __getattr__(self, name: str) -> Asset:
        """Load the asset with the given name."""
        matches = sorted(
            (
                path
                for path in self.path.glob(f"{name}*.*")
                if path.suffix in Asset._class_by_suffix
            ),
            key=lambda p: len(str(p)),
        )

        path = next(iter(matches), None)

        if path is None:
            raise AttributeError(f"Asset not found: {self.path / name}")

        if path.is_dir():
            asset = Assets(path)
        else:
            asset = Asset._class_by_suffix[path.suffix].load(path)

        setattr(self, name, asset)
        self.__annotations__[name] = type(asset)
        return asset

    def __getitem__(self, name: str) -> Asset:
        try:
            return getattr(self, name)
        except AttributeError as attr_error:
            raise KeyError(name) from attr_error

    def __setattr__(self, name: str, value: Any) -> None:
        super().__setattr__(name, value)

    def load_all(self) -> None:
        """Load all the assets in the directory."""
        for child in list(self.path.iterdir()):
            self[child.stem]


class Image(Asset):
    """An image asset."""

    suffixes = {".png", ".jpg", ".jpeg", ".gif"}

    @classmethod
    def load(cls, path: Path) -> pygame.Surface:
        pygame.display.init()
        return pygame.image.load(path)


class Sound(Asset):
    """A sound asset."""

    suffixes = {".wav", ".mp3", ".ogg"}

    @classmethod
    def load(cls, path: Path) -> pygame.mixer.Sound:
        pygame.mixer.init()
        return pygame.mixer.Sound(path)

    @classmethod
    def stream(cls, path: Path) -> Any:
        pygame.mixer.init()
        pygame.mixer.music.load(path)
        return pygame.mixer.music


class Font(Asset):
    """A font asset."""

    suffixes = {".ttf"}

    class _Font:
        def __init__(self, path: Path) -> None:
            self.path = path
            self.cache: dict[int, pygame.font.Font] = {}

        def size(self, size: int = 30) -> pygame.font.Font:
            return self.cache.setdefault(
                size,
                pygame.font.Font(self.path, size),
            )

    @classmethod
    def load(cls, path: Path) -> _Font:
        pygame.font.init()
        return Font._Font(path)


class YAML(Asset):
    """A YAML asset."""

    suffixes = {".yaml", ".yml"}

    @classmethod
    def load(cls, path: Path) -> Any:
        import yaml

        return yaml.safe_load(path.read_text())



import asyncio

import pygame
import pygame.locals as pg

from pygskin import Clock


def run_game(window: pygame.Window, fn, fps: int = 60):
    """
    Run a game loop with the given function.

    Args:
        window (pygame.Window): The window to run the game in.
        fn (Callable): The function to run each frame.
        fps (int): The frames per second to run the game at.
    """
    pygame.init()
    surface = window.get_surface()

    async def _main_loop():
        running = [True]
        quit = running.clear

        while running:
            Clock.tick(fps)
            events = list(pygame.event.get())
            if any(event.type == pg.QUIT for event in events):
                quit()
            fn(surface, events, quit)
            window.flip()
            await asyncio.sleep(0)

    asyncio.run(_main_loop())




from collections.abc import Callable
from contextlib import suppress
from typing import Any

from pygame import Surface
from pygame.event import Event

from pygskin.statemachine import statemachine

ScreenFn = Callable[[Surface, list[Event], Callable], None]
Transition = Callable[[Any], ScreenFn | None]
TransitionTable = dict[ScreenFn, list[Transition]]
EXIT = object()


def screen_manager(transition_table: TransitionTable) -> Callable:
    sm = statemachine(transition_table)

    manager = {
        "screen": next(sm),
    }

    def send(input: Any = EXIT) -> None:
        if input == EXIT:
            manager["exit"]()
            return
        with suppress(StopIteration):
            manager["screen"] = sm.send(input)

    def _screen_manager(surface: Surface, events: list[Event], exit: Callable):
        manager.setdefault("exit", exit)
        manager["screen"](surface, events, send)

    return _screen_manager

